/*
 * FTP directory I/O support routines for PhysicsFS.
 *
 * Please see the file LICENSE.txt in the source's root directory.
 *
 *  This file written by Indy Sams, based on dir.c by Ryan C. Gordon.
 */

/***

FTP Driver version 1.0

This is to be considered alpha quality, it's not well tested but is a starting point for a more
advanced driver (possibly written with the async libCURL multi_* interface).

I'm not that good at C, but at least it is currently working.

The worst part of this current implementation is that files opened with FTP_openRead are fully
downloaded from the FTP server before it returns, so it can be slow with large files or a slow connections.

I've also not done any mutex-work in here, as I'm not familiar enough with PhysFS to know where I need to place them.

Write/Append are currently not supported.

***/

#if (defined PHYSFS_SUPPORTS_FTP)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "physfs.h"

#define __PHYSICSFS_INTERNAL__
#include "physfs_internal.h"

#include "ftpparse/ftpparse.h"

typedef struct {
	void * fp;
	char fn[16];

	void *hHandle;//FTP_HANDLE *
	void *next;
} FTP_FILE_HANDLE;

typedef struct {
	 CURL * cHandle;
	 //void * mHandle;//mutex handle
	 char host[256];
	 int port;
	 char username[128];
	 char password[128];
	 char base_path[256];
	 FTP_FILE_HANDLE * handles;
} FTP_HANDLE;

typedef struct {
	char * ptr;
	size_t len;
} FTP_BUFFER;

static PHYSFS_sint64 FTP_read(fvoid *opaque, void *buffer,
                              PHYSFS_uint32 objSize, PHYSFS_uint32 objCount)
{
	FTP_FILE_HANDLE * h = (FTP_FILE_HANDLE *)opaque;
    PHYSFS_sint64 retval = __PHYSFS_platformRead(h->fp, buffer, objSize, objCount);
    return(retval);
} /* FTP_read */


static PHYSFS_sint64 FTP_write(fvoid *opaque, const void *buffer,
                               PHYSFS_uint32 objSize, PHYSFS_uint32 objCount)
{
	FTP_FILE_HANDLE * h = (FTP_FILE_HANDLE *)opaque;
    PHYSFS_sint64 retval = __PHYSFS_platformWrite(h->fp, buffer, objSize, objCount);
    return(retval);
} /* FTP_write */


static int FTP_eof(fvoid *opaque)
{
	FTP_FILE_HANDLE * h = (FTP_FILE_HANDLE *)opaque;
    return(__PHYSFS_platformEOF(h->fp));
} /* FTP_eof */


static PHYSFS_sint64 FTP_tell(fvoid *opaque)
{
	FTP_FILE_HANDLE * h = (FTP_FILE_HANDLE *)opaque;
    return(__PHYSFS_platformTell(h->fp));
} /* FTP_tell */


static int FTP_seek(fvoid *opaque, PHYSFS_uint64 offset)
{
	FTP_FILE_HANDLE * h = (FTP_FILE_HANDLE *)opaque;
    return(__PHYSFS_platformSeek(h->fp, offset));
} /* FTP_seek */


static PHYSFS_sint64 FTP_fileLength(fvoid *opaque)
{
	FTP_FILE_HANDLE * h = (FTP_FILE_HANDLE *)opaque;
    return(__PHYSFS_platformFileLength(h->fp));
} /* FTP_fileLength */


static int FTP_fileClose(fvoid *opaque)
{
    /*
     * we manually flush the buffer, since that's the place a close will
     *  most likely fail, but that will leave the file handle in an undefined
     *  state if it fails. Flush failures we can recover from.
     */
	FTP_FILE_HANDLE * f = (FTP_FILE_HANDLE *)opaque;
	FTP_HANDLE * h = (FTP_HANDLE *)f->hHandle;
    BAIL_IF_MACRO(!__PHYSFS_platformFlush(f->fp), NULL, 0);
    BAIL_IF_MACRO(!__PHYSFS_platformClose(f->fp), NULL, 0);

	__PHYSFS_platformDelete(f->fn);

	if (h->handles == f) {
		if (f->next) {
			h->handles = f->next;
		} else {
			h->handles = NULL;
		}
	} else {
		FTP_FILE_HANDLE * scan = h->handles, *last=NULL;
		while (scan != f && scan != NULL) {
			last = scan;
			scan = scan->next;
		}
		if (scan == f) {
			last->next = f->next;
		}
	}

    return(1);
} /* FTP_fileClose */


static int FTP_isArchive(const char *filename, int forWriting)
{
	if (forWriting) {
		return 0;
	}

	return !strncmp(filename, "ftp://", 6);
} /* FTP_isArchive */

static void *FTP_openArchive(const char *name, int forWriting)
{
    FTP_HANDLE * retval = NULL;
	char *tmp,*p;

    /* !!! FIXME: when is this not called right before openArchive? */
    BAIL_IF_MACRO(!FTP_isArchive(name, forWriting), ERR_UNSUPPORTED_ARCHIVE, 0);

    retval = allocator.Malloc(sizeof(FTP_HANDLE));
    BAIL_IF_MACRO(retval == NULL, ERR_OUT_OF_MEMORY, NULL);
	memset(retval, 0, sizeof(FTP_HANDLE));
	retval->cHandle = curl_easy_init();
	if (retval->cHandle == NULL) {
		allocator.Free(retval);
		BAIL_MACRO(ERR_GEN_FAILURE, NULL);
	}

	tmp = allocator.Malloc(strlen(name)+1-6);
	strcpy(tmp, name+6);

	//get the path on the FTP server
	p = strchr(tmp, '/');
	if (p) {
		strncpy(retval->base_path, p, sizeof(retval->base_path)-1);
		p[0]=0;
	}

	//determine if we have a username/password
	p = strchr(tmp, '@');
	if (p) {
		p[0]=0;
		strncpy(retval->host, p+1, sizeof(retval->host)-1);
		p = strchr(tmp, ':');
		if (p) {
			p[0]=0;
			strncpy(retval->password, p+1, sizeof(retval->password)-1);
			strncpy(retval->username, tmp, sizeof(retval->username)-1);
		} else {
			strncpy(retval->username, tmp, sizeof(retval->username)-1);
		}
	} else {
		strncpy(retval->host, tmp, sizeof(retval->host)-1);
	}

	p = strchr(retval->host, ':');
	if (p) {
		p[0]=0;
		retval->port = atoi(p+1);
	}

	allocator.Free(tmp);

	if (!retval->port) { retval->port = 21; }
	if (!strlen(retval->username)) { strcpy(retval->username, "anonymous"); }
	if (!strlen(retval->password)) { strcpy(retval->password, "fake@email.com"); }
	if (!strlen(retval->base_path)) { strcpy(retval->base_path, "/"); }
	if (retval->base_path[strlen(retval->base_path)-1] != '/') {
		strcat(retval->base_path, "/");
	}
    return(retval);
} /* FTP_openArchive */

static size_t FTP_enumerateFiles_cb( void *ptr, size_t size, size_t nmemb, void *stream) {
	FTP_BUFFER * buf = stream;
	size_t len = (size*nmemb);
	if (len > 0) {
		buf->ptr = allocator.Realloc(buf->ptr, buf->len+len);
		memcpy(buf->ptr+buf->len, ptr, len);
		buf->len += len;
	}
	return len;
}

static void FTP_enumerateFiles(dvoid *opaque, const char *dname,
                               int omitSymLinks, PHYSFS_EnumFilesCallback cb,
                               const char *origdir, void *callbackdata)
{
    FTP_HANDLE * h = opaque;
	CURLcode res;
	char buf[2048];
	FTP_BUFFER fb;
	memset(&fb, 0, sizeof(fb));
	sprintf(buf, "ftp://%s:%s@%s:%d%s%s", h->username, h->password, h->host, h->port, h->base_path, dname);

	if (buf[strlen(buf)-1] != '/') { strcat(buf, "/"); }
	
	//__PHYSFS_platformGrabMutex(h->mHandle);
	curl_easy_setopt(h->cHandle, CURLOPT_URL, buf);
	curl_easy_setopt(h->cHandle, CURLOPT_FTP_FILEMETHOD, CURLFTPMETHOD_SINGLECWD);//CURLFTPMETHOD_MULTICWD
	curl_easy_setopt(h->cHandle, CURLOPT_CONNECTTIMEOUT, 10);
	curl_easy_setopt(h->cHandle, CURLOPT_WRITEFUNCTION, FTP_enumerateFiles_cb);
	curl_easy_setopt(h->cHandle, CURLOPT_WRITEDATA, &fb);
	curl_easy_setopt(h->cHandle, CURLOPT_CUSTOMREQUEST, "LIST -l");

	res = curl_easy_perform(h->cHandle);
	if (res != CURLE_OK) {
		//retry once
		curl_easy_setopt(h->cHandle, CURLOPT_FTP_FILEMETHOD, CURLFTPMETHOD_MULTICWD);
		res = curl_easy_perform(h->cHandle);
	}
	curl_easy_reset(h->cHandle);
	//__PHYSFS_platformReleaseMutex(h->mHandle);
	if (res == CURLE_OK) {
		char * p;
		fb.ptr = allocator.Realloc(fb.ptr, fb.len+1);
		fb.ptr[fb.len] = 0;

		p = fb.ptr;
		while (p) {
			struct ftpparse fp;
			char * q = strchr(p, '\n');
			if (q) {
				q[0]=0;
				q++;
			}
			while (p[strlen(p)-1] == '\r') { p[strlen(p)-1]=0; }
			memset(&fp, 0, sizeof(fp));
			if (ftpparse(&fp, p, strlen(p)) == 1) {
				fp.name[fp.namelen]=0;
				cb(callbackdata, origdir, fp.name);
				//printf("line: %s\n", p);
			}
			p = q;
		}
		//printf("list: %s\n", fb.ptr);
	}
	if (fb.ptr) { allocator.Free(fb.ptr); }
} /* FTP_enumerateFiles */


static int FTP_exists(dvoid *opaque, const char *name)
{
     FTP_HANDLE * h = opaque;
	CURLcode res;
	char buf[2048];
	FTP_BUFFER fb;
	char *dir, *fn, *p;
    int retval = 0, len=0;
	memset(&fb, 0, sizeof(fb));
	if ((p=strrchr(name, '/')) != NULL) {
		fn = p+1;
		len=p-name+1;
		dir = allocator.Malloc(len+1);
		strncpy(dir, name, len);
		dir[len]=0;
		sprintf(buf, "ftp://%s:%s@%s:%d%s%s", h->username, h->password, h->host, h->port, h->base_path, dir);
	} else {
		sprintf(buf, "ftp://%s:%s@%s:%d%s", h->username, h->password, h->host, h->port, h->base_path);
		dir = NULL;
		fn = (char *)name;
	}
	memset(&fb, 0, sizeof(fb));
	//sprintf(buf, "ftp://%s:%s@%s:%d%s%s", h->username, h->password, h->host, h->port, h->base_path, name);

	//if (buf[strlen(buf)-1] != '/') { strcat(buf, "/"); }
	
	//__PHYSFS_platformGrabMutex(h->mHandle);
	curl_easy_setopt(h->cHandle, CURLOPT_URL, buf);
	curl_easy_setopt(h->cHandle, CURLOPT_FTP_FILEMETHOD, CURLFTPMETHOD_SINGLECWD);//CURLFTPMETHOD_MULTICWD
	curl_easy_setopt(h->cHandle, CURLOPT_CONNECTTIMEOUT, 10);
	curl_easy_setopt(h->cHandle, CURLOPT_WRITEFUNCTION, FTP_enumerateFiles_cb);
	curl_easy_setopt(h->cHandle, CURLOPT_WRITEDATA, &fb);
	curl_easy_setopt(h->cHandle, CURLOPT_CUSTOMREQUEST, "LIST -l");

	res = curl_easy_perform(h->cHandle);
	if (res != CURLE_OK) {
		//retry once
		curl_easy_setopt(h->cHandle, CURLOPT_FTP_FILEMETHOD, CURLFTPMETHOD_MULTICWD);
		res = curl_easy_perform(h->cHandle);
	}
	curl_easy_reset(h->cHandle);
	//__PHYSFS_platformReleaseMutex(h->mHandle);
	if (res == CURLE_OK) {
		char * p;
		fb.ptr = allocator.Realloc(fb.ptr, fb.len+1);
		fb.ptr[fb.len] = 0;

		p = fb.ptr;
		while (p) {
			struct ftpparse fp;
			char * q = strchr(p, '\n');
			if (q) {
				q[0]=0;
				q++;
			}
			while (p[strlen(p)-1] == '\r') { p[strlen(p)-1]=0; }
			memset(&fp, 0, sizeof(fp));
			if (ftpparse(&fp, p, strlen(p)) == 1) {
				fp.name[fp.namelen]=0;
				if (!stricmp(fp.name, fn)) {
					retval = 1;
					q=NULL;
				}
				//cb(callbackdata, origdir, fp.name);
				//printf("line: %s\n", p);
			}
			p = q;
		}
		//printf("list: %s\n", fb.ptr);
	}
	if (fb.ptr) { allocator.Free(fb.ptr); }
	if (dir) { allocator.Free(dir); }
    return(retval);
} /* FTP_exists */


static int FTP_isDirectory(dvoid *opaque, const char *name, int *fileExists)
{
	//this is a bit of a hack-job, but it was the only way I could get it to work
    FTP_HANDLE * h = opaque;
	CURLcode res;
	char buf[2048];
	FTP_BUFFER fb;
	int retval = 0;

	memset(&fb, 0, sizeof(fb));
	sprintf(buf, "ftp://%s:%s@%s:%d%s%s", h->username, h->password, h->host, h->port, h->base_path, name);

	if (buf[strlen(buf)-1] != '/') { strcat(buf, "/"); }
	
	//__PHYSFS_platformGrabMutex(h->mHandle);
	curl_easy_setopt(h->cHandle, CURLOPT_URL, buf);
	curl_easy_setopt(h->cHandle, CURLOPT_FTP_FILEMETHOD, CURLFTPMETHOD_SINGLECWD);//CURLFTPMETHOD_MULTICWD
	curl_easy_setopt(h->cHandle, CURLOPT_CONNECTTIMEOUT, 10);
	curl_easy_setopt(h->cHandle, CURLOPT_WRITEFUNCTION, FTP_enumerateFiles_cb);
	curl_easy_setopt(h->cHandle, CURLOPT_WRITEDATA, &fb);
	curl_easy_setopt(h->cHandle, CURLOPT_CUSTOMREQUEST, "LIST -l");

	res = curl_easy_perform(h->cHandle);
	if (res != CURLE_OK) {
		//retry once
		curl_easy_setopt(h->cHandle, CURLOPT_FTP_FILEMETHOD, CURLFTPMETHOD_MULTICWD);
		res = curl_easy_perform(h->cHandle);
	}
	curl_easy_reset(h->cHandle);
	//__PHYSFS_platformReleaseMutex(h->mHandle);
	if (res == CURLE_OK) {
		//if CURL was successful, then we should have a file list. I would parse it to be sure but an empty dir might return a false negative that way
		retval = 1;
	}
	if (fb.ptr) { allocator.Free(fb.ptr); }
	if (fileExists) { *fileExists = retval; }
	return (retval);
} /* FTP_isDirectory */


static int FTP_isSymLink(dvoid *opaque, const char *name, int *fileExists)
{
	if (fileExists) { *fileExists = FTP_exists(opaque, name); }
    return(0);
} /* FTP_isSymLink */


static PHYSFS_sint64 FTP_getLastModTime(dvoid *opaque,
                                        const char *name,
                                        int *fileExists)
{
    FTP_HANDLE * h = opaque;
	CURLcode res;
	char buf[2048];
	FTP_BUFFER fb;
	char *dir, *fn, *p;
    PHYSFS_sint64 retval = -1;
	int len=0;
	memset(&fb, 0, sizeof(fb));
	if ((p=strchr(name, '/')) != NULL) {
		fn = p+1;
		len=p-name+1;
		dir = allocator.Malloc(len+1);
		strncpy(dir, name, len);
		dir[len]=0;
		sprintf(buf, "ftp://%s:%s@%s:%d%s%s", h->username, h->password, h->host, h->port, h->base_path, dir);
	} else {
		sprintf(buf, "ftp://%s:%s@%s:%d%s", h->username, h->password, h->host, h->port, h->base_path);
		dir = NULL;
		fn = (char *)name;
	}
	memset(&fb, 0, sizeof(fb));
	//sprintf(buf, "ftp://%s:%s@%s:%d%s%s", h->username, h->password, h->host, h->port, h->base_path, name);

	//if (buf[strlen(buf)-1] != '/') { strcat(buf, "/"); }
	
	//__PHYSFS_platformGrabMutex(h->mHandle);
	curl_easy_setopt(h->cHandle, CURLOPT_URL, buf);
	curl_easy_setopt(h->cHandle, CURLOPT_FTP_FILEMETHOD, CURLFTPMETHOD_SINGLECWD);//CURLFTPMETHOD_MULTICWD
	curl_easy_setopt(h->cHandle, CURLOPT_CONNECTTIMEOUT, 10);
	curl_easy_setopt(h->cHandle, CURLOPT_WRITEFUNCTION, FTP_enumerateFiles_cb);
	curl_easy_setopt(h->cHandle, CURLOPT_WRITEDATA, &fb);
	curl_easy_setopt(h->cHandle, CURLOPT_CUSTOMREQUEST, "LIST -l");

	res = curl_easy_perform(h->cHandle);
	if (res != CURLE_OK) {
		//retry once
		curl_easy_setopt(h->cHandle, CURLOPT_FTP_FILEMETHOD, CURLFTPMETHOD_MULTICWD);
		res = curl_easy_perform(h->cHandle);
	}
	curl_easy_reset(h->cHandle);
	//__PHYSFS_platformReleaseMutex(h->mHandle);
	if (res == CURLE_OK) {
		char * p;
		fb.ptr = allocator.Realloc(fb.ptr, fb.len+1);
		fb.ptr[fb.len] = 0;

		p = fb.ptr;
		while (p) {
			struct ftpparse fp;
			char * q = strchr(p, '\n');
			if (q) {
				q[0]=0;
				q++;
			}
			while (p[strlen(p)-1] == '\r') { p[strlen(p)-1]=0; }
			memset(&fp, 0, sizeof(fp));
			if (ftpparse(&fp, p, strlen(p)) == 1) {
				fp.name[fp.namelen]=0;
				if (!stricmp(fp.name, fn)) {
					retval = fp.mtime;
					q=NULL;
				}
				//cb(callbackdata, origdir, fp.name);
				//printf("line: %s\n", p);
			}
			p = q;
		}
		//printf("list: %s\n", fb.ptr);
	}
	if (fb.ptr) { allocator.Free(fb.ptr); }
	if (dir) { allocator.Free(dir); }
	if (fileExists) {
		*fileExists = retval == -1 ? FTP_exists(opaque, name):1;
	}

    return(retval);
} /* FTP_getLastModTime */

static size_t FTP_openRead_cb( void *ptr, size_t size, size_t nmemb, void *stream) {
	return __PHYSFS_platformWrite(stream, ptr, size, nmemb);
}

static fvoid *FTP_openRead(dvoid *opaque, const char *fnm, int *exist)
{
	FTP_FILE_HANDLE *retval, *scan;
	void * fp;
	char fn[16];
	char buf[2048];
	int i=0;
    FTP_HANDLE * h = opaque;
	CURLcode res;

	if (exist) { *exist = FTP_exists(opaque, fnm); }
	if (exist && !*exist) { return (0); }

	sprintf(fn, "ftp.%d.tmp", i);
	while (__PHYSFS_platformExists(fn)) {
		i++;
		sprintf(fn, "ftp.%d.tmp", i);
	}

	fp = __PHYSFS_platformOpenWrite(fn);
	if (fp == NULL) { return (0); }

	sprintf(buf, "ftp://%s:%s@%s:%d%s%s", h->username, h->password, h->host, h->port, h->base_path, fnm);

	curl_easy_setopt(h->cHandle, CURLOPT_URL, buf);
	curl_easy_setopt(h->cHandle, CURLOPT_FTP_FILEMETHOD, CURLFTPMETHOD_SINGLECWD);//CURLFTPMETHOD_MULTICWD
	curl_easy_setopt(h->cHandle, CURLOPT_CONNECTTIMEOUT, 10);
	curl_easy_setopt(h->cHandle, CURLOPT_WRITEFUNCTION, FTP_openRead_cb);
	curl_easy_setopt(h->cHandle, CURLOPT_WRITEDATA, fp);

	res = curl_easy_perform(h->cHandle);
	if (res != CURLE_OK) {
		//retry once
		curl_easy_setopt(h->cHandle, CURLOPT_FTP_FILEMETHOD, CURLFTPMETHOD_MULTICWD);
		res = curl_easy_perform(h->cHandle);
	}
	curl_easy_reset(h->cHandle);
	__PHYSFS_platformClose(fp);
	//__PHYSFS_platformReleaseMutex(h->mHandle);
	if (res == CURLE_OK) {
		retval = allocator.Malloc(sizeof(FTP_FILE_HANDLE));
		memset(retval, 0, sizeof(FTP_FILE_HANDLE));
		retval->fp = __PHYSFS_platformOpenRead(fn);
		retval->hHandle = h;
		strcpy(retval->fn, fn);
		if (h->handles == NULL) {
			h->handles = retval;
		} else {
			scan = h->handles;
			while (scan->next) {
				scan = scan->next;
			}
			scan->next = retval;
		}
	}
    return retval;
} /* FTP_openRead */


static fvoid *FTP_openWrite(dvoid *opaque, const char *filename)
{
	return 0;
    //return(doOpen(opaque, filename, __PHYSFS_platformOpenWrite, NULL));
} /* FTP_openWrite */


static fvoid *FTP_openAppend(dvoid *opaque, const char *filename)
{
	return 0;
    //return(doOpen(opaque, filename, __PHYSFS_platformOpenAppend, NULL));
} /* FTP_openAppend */


static int FTP_remove(dvoid *opaque, const char *name)
{
    return(0);
} /* FTP_remove */


static int FTP_mkdir(dvoid *opaque, const char *name)
{
    return(0);
} /* FTP_mkdir */


static void FTP_dirClose(dvoid *opaque)
{
    FTP_HANDLE * retval = opaque;

	if (opaque == NULL) {
		__PHYSFS_setError(ERR_NOT_AN_ARCHIVE);
		return;
	}

	if (retval->handles) {
		fprintf(stdout, "ERROR: Not all FTP files have been closed!!!\n");
	}
	if (retval->cHandle != NULL) {
		curl_easy_cleanup(retval->cHandle);
	}
    allocator.Free(opaque);
} /* FTP_dirClose */



const PHYSFS_ArchiveInfo __PHYSFS_ArchiveInfo_FTP =
{
    "",
    "FTP Server Access",
    "Indy Sams <indy@shoutirc.com>, based on dir.c by Ryan C. Gordon <icculus@icculus.org>",
    "http://www.shoutirc.com",
};

const PHYSFS_Archiver __PHYSFS_Archiver_FTP =
{
    &__PHYSFS_ArchiveInfo_FTP,
    FTP_isArchive,          /* isArchive() method      */
    FTP_openArchive,        /* openArchive() method    */
    FTP_enumerateFiles,     /* enumerateFiles() method */
    FTP_exists,             /* exists() method         */
    FTP_isDirectory,        /* isDirectory() method    */
    FTP_isSymLink,          /* isSymLink() method      */
    FTP_getLastModTime,     /* getLastModTime() method */
    FTP_openRead,           /* openRead() method       */
    FTP_openWrite,          /* openWrite() method      */
    FTP_openAppend,         /* openAppend() method     */
    FTP_remove,             /* remove() method         */
    FTP_mkdir,              /* mkdir() method          */
    FTP_dirClose,           /* dirClose() method       */
    FTP_read,               /* read() method           */
    FTP_write,              /* write() method          */
    FTP_eof,                /* eof() method            */
    FTP_tell,               /* tell() method           */
    FTP_seek,               /* seek() method           */
    FTP_fileLength,         /* fileLength() method     */
    FTP_fileClose           /* fileClose() method      */
};

/* end of ftp.c ... */
#endif
